use super::Keyfork;
use clap::{Parser, Subcommand};

use keyfork_derive_openpgp::openpgp::{
    armor::{Kind, Writer},
    packet::UserID,
    serialize::Marshal,
    types::KeyFlags,
};
use keyfork_derive_util::{
    request::{DerivationAlgorithm, DerivationRequest, DerivationResponse},
    DerivationIndex, DerivationPath,
};
use keyforkd_client::Client;

type Result<T, E = Box<dyn std::error::Error>> = std::result::Result<T, E>;

#[derive(Subcommand, Clone, Debug)]
pub enum DeriveSubcommands {
    /// Derive an OpenPGP certificate.
    #[command(name = "openpgp")]
    OpenPGP {
        /// Default User ID for the certificate.
        user_id: String,
    },
}

impl DeriveSubcommands {
    fn handle(&self, account: DerivationIndex) -> Result<()> {
        match self {
            DeriveSubcommands::OpenPGP { user_id } => {
                let mut pgp_u32 = [0u8; 4];
                pgp_u32[1..].copy_from_slice(&"pgp".bytes().collect::<Vec<u8>>());
                let chain = DerivationIndex::new(u32::from_be_bytes(pgp_u32), true)?;
                let path = DerivationPath::default()
                    .chain_push(chain)
                    .chain_push(account);
                // TODO: should this be customizable?
                let subkeys = vec![
                    KeyFlags::empty().set_certification(),
                    KeyFlags::empty().set_signing(),
                    KeyFlags::empty()
                        .set_transport_encryption()
                        .set_storage_encryption(),
                    KeyFlags::empty().set_authentication(),
                ];
                let request = DerivationRequest::new(DerivationAlgorithm::Ed25519, &path);
                let derived_data: DerivationResponse = Client::discover_socket()?
                    .request(&request.into())?
                    .try_into()?;
                let default_userid = UserID::from(user_id.as_str());
                let cert = keyfork_derive_openpgp::derive(derived_data, &subkeys, &default_userid)?;

                let mut w = Writer::new(std::io::stdout(), Kind::SecretKey)?;

                for packet in cert.into_packets() {
                    packet.serialize(&mut w)?;
                }

                w.finalize()?;
            }
        }

        Ok(())
    }
}

#[derive(Parser, Debug, Clone)]
pub struct Derive {
    #[command(subcommand)]
    command: DeriveSubcommands,

    /// Account ID. Required for all derivations.
    #[arg(long, global = true, default_value = "0")]
    account_id: u32,
}

impl Derive {
    pub fn handle(&self, _k: &Keyfork) -> Result<()> {
        let account = DerivationIndex::new(self.account_id, true)?;
        self.command.handle(account)
    }
}
