use super::Keyfork;
use clap::{Parser, Subcommand};
use std::path::PathBuf;

use keyfork_mnemonic_util::Mnemonic;
use keyfork_shard::{
    openpgp::{combine, discover_certs, parse_messages},
    remote_decrypt,
};

type Result<T, E = Box<dyn std::error::Error>> = std::result::Result<T, E>;

#[derive(Subcommand, Clone, Debug)]
pub enum RecoverSubcommands {
    /// Decrypt a shard file using keys available on the local system.
    Shard {
        shard_file: PathBuf,
        key_discovery: Option<PathBuf>,
    },

    /// Combine remotely decrypted shards.
    RemoteShard {},
}

impl RecoverSubcommands {
    fn handle(&self) -> Result<Vec<u8>> {
        match self {
            RecoverSubcommands::Shard {
                shard_file,
                key_discovery,
            } => {
                let content = std::fs::read_to_string(shard_file)?;
                if content.contains("BEGIN PGP MESSAGE") {
                    let certs = key_discovery
                        .as_ref()
                        .map(discover_certs)
                        .transpose()?
                        .unwrap_or(vec![]);
                    let mut messages = parse_messages(content.as_bytes())?;
                    let metadata = messages.pop_front().expect("any pgp encrypted message");
                    let mut seed = vec![];
                    combine(certs, &metadata, messages.into(), &mut seed)?;
                    Ok(seed)
                } else {
                    panic!("unknown format of shard file");
                }
            }
            RecoverSubcommands::RemoteShard {} => {
                let mut seed = vec![];
                remote_decrypt(&mut seed)?;
                Ok(seed)
            }
        }
    }
}

#[derive(Parser, Debug, Clone)]
pub struct Recover {
    #[command(subcommand)]
    command: RecoverSubcommands,
}

impl Recover {
    pub fn handle(&self, _k: &Keyfork) -> Result<()> {
        let seed = self.command.handle()?;
        let mnemonic = Mnemonic::from_entropy(&seed, Default::default())?;
        tokio::runtime::Builder::new_multi_thread()
            .enable_all()
            .build()
            .unwrap()
            .block_on(async {
                keyforkd::setup_registry();
                keyforkd::start_and_run_server(mnemonic).await
            })
    }
}
